// (C) 2017-2020 by folkert van heusden, released under AGPL v3.0
#include <dlfcn.h>
#include <libconfig.h++>
#include <set>
#include <string>
#include <string.h>
#include <vector>

using namespace libconfig;

#include "cfg.h"
#include "config.h"
#include "error.h"
#include "utils.h"
#include "source.h"
#include "source_v4l.h"
#include "source_http_mjpeg.h"
#include "source_http_jpeg.h"
#include "source_stream.h"
#include "source_plugin.h"
#include "source_gstreamer.h"
#include "source_pixelflood.h"
#include "http_server.h"
#include "motion_trigger.h"
#include "motion_trigger_generic.h"
#include "motion_trigger_other_source.h"
#include "target.h"
#ifdef WITH_GWAVI
#include "target_avi.h"
#endif
#include "target_ffmpeg.h"
#include "target_jpeg.h"
#include "target_plugin.h"
#include "target_extpipe.h"
#include "target_vnc.h"
#include "target_pixelflood.h"
#include "target_new_source.h"
#include "filter.h"
#include "filter_mirror_v.h"
#include "filter_mirror_h.h"
#include "filter_noise_neighavg.h"
#include "filter_apply_mask.h"
#include "filter_add_text.h"
#include "filter_add_scaled_text.h"
#include "filter_grayscale.h"
#include "filter_boost_contrast.h"
#include "filter_marker_simple.h"
#include "filter_average.h"
#include "filter_median.h"
#include "filter_plugin_frei0r.h"
#include "v4l2_loopback.h"
#include "filter_overlay.h"
#include "picio.h"
#include "filter_plugin.h"
#include "log.h"
#include "cfg.h"
#include "resize_cairo.h"
#include "resize_crop.h"
#include "resize_fine.h"
#include "filter_motion_only.h"
#include "selection_mask.h"
#include "cleaner.h"
#include "view_html_grid.h"
#include "view_ss.h"
#include "view_html_all.h"
#include "view_all.h"
#include "view_pip.h"
#include "source_delay.h"
#include "http_auth.h"
#include "http_auth_pam.h"

void find_by_id(const configuration_t *const cfg, const std::string & id, instance_t **inst, interface **i)
{
	*inst = nullptr;
	*i = nullptr;

	for(instance_t * cur_inst : cfg -> instances) {
		for(interface *cur_i : cur_inst -> interfaces) {
			if (cur_i -> get_id() == id) {
				*inst = cur_inst;
				*i = cur_i;
			}
		}
	}
}

source *find_source(instance_t *const inst)
{
	if (!inst)
		return nullptr;

	for(interface *i : inst -> interfaces) {
		if (i -> get_class_type() == CT_SOURCE)
			return (source *)i;
	}

	return nullptr;
}

std::vector<interface *> find_motion_triggers(instance_t *const inst)
{
	std::vector<interface *> out;

	if (inst) {
		for(interface *i : inst -> interfaces) {
			if (i -> get_class_type() == CT_MOTIONTRIGGER)
				out.push_back(i);
		}
	}

	return out;
}

instance_t *find_instance_by_interface(const configuration_t *const cfg, const interface *i_f)
{
	for(instance_t * inst : cfg -> instances) {
		for(interface *cur : inst -> interfaces) {
			if (cur == i_f)
				return inst;
		}
	}

	return nullptr;
}

instance_t *find_instance_by_name(const configuration_t *const cfg, const std::string & name)
{
	for(instance_t * inst : cfg -> instances) {
		if (inst -> name == name)
			return inst;
	}

	return nullptr;
}

interface *find_interface_by_id(instance_t *const inst, const std::string & id)
{
	if (!inst)
		return nullptr;

	for(interface *i : inst -> interfaces) {
		std::string id_int = i -> get_id();

		if (id == id_int)
			return i;
	}

	return nullptr;
}

interface *find_interface_by_id(configuration_t *const cfg, const std::string & id)
{
	for(instance_t * inst : cfg -> instances) {
		interface *i = find_interface_by_id(inst, id);
		if (i)
			return i;
	}

	return nullptr;
}

bool check_for_motion(instance_t *const inst)
{
	for(interface *i : inst -> interfaces) {
		if (i -> get_class_type() == CT_MOTIONTRIGGER && ((motion_trigger *)i) -> check_motion())
			return true;
	}

	return false;
}

std::string cfg_str(const Config & cfg, const char *const key, const char *descr, const bool optional, const std::string & def)
{
	try {
		return (const char *)cfg.lookup(key);
	}
	catch(const SettingNotFoundException &nfex) {
		if (!optional)
			error_exit(false, "\"%s\" not found (%s)", key, descr);
	}

	log(LL_DEBUG, "\"%s\" not found (%s), assuming default (%s)", key, descr, def.c_str());

	return def; // field is optional
}

std::string cfg_str(const Setting & cfg, const char *const key, const char *descr, const bool optional, const std::string & def)
{
	std::string v = def;

	try {
		v = (const char *)cfg.lookup(key);
	}
	catch(const SettingNotFoundException &nfex) {
		if (!optional)
			error_exit(false, "\"%s\" not found (%s)", key, descr);

		log(LL_DEBUG, "\"%s\" not found (%s), assuming default (%s)", key, descr, def.c_str());
	}

	catch(const SettingTypeException & ste) {
		error_exit(false, "Expected a string value for \"%s\" (%s) at line %d but got something else", key, descr, cfg.getSourceLine());
	}

	return v;
}

double cfg_float(const Setting & cfg, const char *const key, const char *descr, const bool optional, const double def=-1.0, std::map<std::string, parameter *> *const dp = nullptr)
{
	double v = def;

	try {
		v = cfg.lookup(key);
	}
	catch(const SettingNotFoundException &nfex) {
		if (!optional)
			error_exit(false, "\"%s\" not found (%s)", key, descr);

		log(LL_DEBUG, "\"%s\" not found (%s), assuming default (%f)", key, descr, def);
	}

	catch(const SettingTypeException & ste) {
		error_exit(false, "Expected a float value for \"%s\" (%s) at line %d but got something else (did you forget to add \".0\"?)", key, descr, cfg.getSourceLine());
	}

	if (dp)
		parameter::add_value(*dp, key, descr, v);

	return v;
}

int cfg_int(const Setting & cfg, const char *const key, const char *descr, const bool optional, const int def=-1, std::map<std::string, parameter *> *const dp = nullptr)
{
	int v = def;

	try {
		v = cfg.lookup(key);
	}
	catch(const SettingNotFoundException &nfex) {
		if (!optional)
			error_exit(false, "\"%s\" not found (%s)", key, descr);

		log(LL_DEBUG, "\"%s\" not found (%s), assuming default (%f)", key, descr, def);
	}

	catch(const SettingTypeException & ste) {
		error_exit(false, "Expected an int value for \"%s\" (%s) at line %d but got something else", key, descr, cfg.getSourceLine());
	}

	if (dp)
		parameter::add_value(*dp, key, descr, v);

	return v;
}

int cfg_bool(const Config & cfg, const char *const key, const char *descr, const bool optional, const bool def=false)
{
	bool v = def;

	try {
		v = cfg.lookup(key);
	}
	catch(const SettingNotFoundException &nfex) {
		if (!optional)
			error_exit(false, "\"%s\" not found (%s)", key, descr);

		log(LL_DEBUG, "\"%s\" not found (%s), assuming default (%f)", key, descr, def);
	}

	catch(const SettingTypeException & ste) {
		error_exit(false, "Expected a boolean value for \"%s\" (%s) but got something else", key, descr);
	}

	return v;
}

int cfg_bool(const Setting & cfg, const char *const key, const char *descr, const bool optional, const bool def=false)
{
	bool v = def;

	try {
		v = cfg.lookup(key);
	}
	catch(const SettingNotFoundException &nfex) {
		if (!optional)
			error_exit(false, "\"%s\" not found (%s)", key, descr);

		log(LL_DEBUG, "\"%s\" not found (%s), assuming default (%f)", key, descr, def);
	}

	catch(const SettingTypeException & ste) {
		error_exit(false, "Expected a boolean value for \"%s\" (%s) at line %d but got something else", key, descr, cfg.getSourceLine());
	}

	return v;
}

pos_t pos_to_pos(const std::string & s_position)
{
	pos_t pos{ center_center, 0, 0 };

	if (s_position == "upper-left")
		pos.type = upper_left;
	else if (s_position == "upper-center")
		pos.type = upper_center;
	else if (s_position == "upper-right")
		pos.type = upper_right;
	else if (s_position == "center-left")
		pos.type = center_left;
	else if (s_position == "center-center")
		pos.type = center_center;
	else if (s_position == "center-right")
		pos.type = center_right;
	else if (s_position == "lower-left")
		pos.type = lower_left;
	else if (s_position == "lower-center")
		pos.type = lower_center;
	else if (s_position == "lower-right")
		pos.type = lower_right;
	else if (s_position.substr(0, 3) == "xy:") {
		pos.type = xy;
		std::vector<std::string> *parts = split(s_position.substr(4), ",");
		if (parts->size() != 2)
			error_exit(false, "Position \"xy\" must be in format \"xy:...x...,...y...\"");

		pos.x = atoi(parts->at(0).c_str());
		pos.y = atoi(parts->at(1).c_str());
		delete parts;
	}
	else if (s_position == "none")
		pos.type = none;
	else
		error_exit(false, "(text-)position %s is not understood", s_position.c_str());

	return pos;
}

pixelflood_protocol_t cfg_convert_pp(const std::string & pp_str)
{
	pixelflood_protocol_t pp = pixelflood_protocol_t(-1);
	if (pp_str == "tcp-txt")
		pp = PP_TCP_TXT;
	else if (pp_str == "udp-bin")
		pp = PP_UDP_BIN;
	else
		error_exit(false, "Pixelflut format \"%s\" is unknown", pp_str.c_str());

	return pp;
}

void add_filters(std::vector<filter *> *af, const std::vector<filter *> *const in)
{
	for(filter *f : *in)
		af -> push_back(f);
}

bool find_interval_or_fps(const Setting & cfg, double *const interval, const std::string & fps_name, double *const fps)
{
	bool have_interval = cfg.lookupValue("interval", *interval);
	bool have_fps = cfg.lookupValue(fps_name.c_str(), *fps);

	if (!have_interval && !have_fps)
		return false;

	if (have_interval && have_fps)
		return false;

	if (have_interval) {
		if (*interval == 0)
			return false;

		if (*interval < 0)
			*fps = -1.0;
		else
			*fps = 1.0 / *interval;
	}
	else {
		if (*fps == 0)
			return false;

		if (*fps <= 0)
			*interval = -1.0;
		else
			*interval = 1.0 / *fps;
	}

	return true;
}

std::vector<filter *> *load_filters(configuration_t *const cfg, const Setting & in, resize *const r, meta *const m)
{
	std::vector<filter *> *const filters = new std::vector<filter *>();

	size_t n_f = in.getLength();
	log(LL_DEBUG, " %zu filters", n_f);

	for(size_t i=0; i<n_f; i++) {
		const Setting & ae = in[i];

		std::string s_type = cfg_str(ae, "type", "filter-type (h-mirror, marker, etc)", false, "");
		if (s_type == "h-mirror")
			filters -> push_back(new filter_mirror_h());
		else if (s_type == "v-mirror")
			filters -> push_back(new filter_mirror_v());
		else if (s_type == "neighbour-noise-filter")
			filters -> push_back(new filter_noise_neighavg());
		else if (s_type == "grayscale")
			filters -> push_back(new filter_grayscale());
		else if (s_type == "filter-plugin") {
			std::string file = cfg_str(ae, "file", "filename of filter plugin", false, "");
			std::string par = cfg_str(ae, "par", "parameter for filter plugin", true, "");

			filters -> push_back(new filter_plugin(file, par));
		}
		else if (s_type == "filter-plugin-frei0r") {
			std::string file = cfg_str(ae, "file", "filename of frei0r filter plugin", false, "");
			std::string pars = cfg_str(ae, "pars", "parameters for frei0r filter plugin (seperated by spaces, k=v pairs)", true, "");

			filters -> push_back(new filter_plugin_frei0r(file, pars));
		}
		else if (s_type == "marker") {
			std::string motion_source = cfg_str(ae, "motion-source", "source to monitor, leave empty for monitoring by itself (less efficient)", true, "");
			instance_t *const i = motion_source.empty() ? nullptr : (instance_t *)find_instance_by_name(cfg, motion_source);

			std::string s_type = cfg_str(ae, "marker-type", "marker type, e.g. box, cross, circle or border", true, "box");

			sm_type_t st = t_box;
			if (s_type == "box")
				st = t_box;
			else if (s_type == "cross")
				st = t_cross;
			else if (s_type == "circle")
				st = t_circle;
			else if (s_type == "border")
				st = t_border;

			sm_mode_t sm = m_red;
			if (st != t_border) {
				std::string s_mode = cfg_str(ae, "mode", "marker mode, e.g. red, red-invert or invert", false, "red");

				if (s_mode == "red")
					sm = m_red;
				else if (s_mode == "red-invert")
					sm = m_red_invert;
				else if (s_mode == "invert")
					sm = m_invert;
			}

			int thick = cfg_int(ae, "thick", "thickness of the marker", true, 2);

			std::string selection_bitmap = cfg_str(ae, "selection-bitmap", "bitmaps indicating which pixels to look at. must be same size as webcam image and must be a .pbm-file. leave empty to disable.", true, "");
			selection_mask *psm = selection_bitmap.empty() ? nullptr : new selection_mask(r, selection_bitmap);
			int noise_level = cfg_int(ae, "noise-factor", "at what difference levell is the pixel considered to be changed", true, 32);

			double pixels_changed_perctange = i ? -1.0 : cfg_float(ae, "pixels-changed-percentage", "what %% of pixels need to be changed before the marker is drawn", false, 1.0);

			filters -> push_back(new filter_marker_simple(i, sm, st, psm, m, noise_level, pixels_changed_perctange, thick));
		}
		else if (s_type == "apply-mask") {
			std::string selection_bitmap = cfg_str(ae, "selection-bitmap", "bitmaps indicating which pixels to look at. must be same size as webcam image and must be a .pbm-file. leave empty to disable.", false, "");
			selection_mask *sm = selection_bitmap.empty() ? nullptr : new selection_mask(r, selection_bitmap);

			filters -> push_back(new filter_apply_mask(sm));
		}
		else if (s_type == "motion-only") {
			std::string selection_bitmap = cfg_str(ae, "selection-bitmap", "bitmaps indicating which pixels to look at. must be same size as webcam image and must be a .pbm-file. leave empty to disable.", true, "");
			selection_mask *sm = selection_bitmap.empty() ? nullptr : new selection_mask(r, selection_bitmap);
			int noise_level = cfg_int(ae, "noise-factor", "at what difference levell is the pixel considered to be changed", true, 32);
			bool diff_only = cfg_bool(ae, "diff-only", "show difference in pixel value, not original picture", true, false);

			filters -> push_back(new filter_motion_only(sm, noise_level, diff_only));
		}
		else if (s_type == "boost-contrast")
			filters -> push_back(new filter_boost_contrast());
		else if (s_type == "overlay") {
			std::string s_pic = cfg_str(ae, "picture", "PNG to overlay (with alpha channel!)", false, "");
			int x = cfg_int(ae, "x", "x-coordinate of overlay", false, 0);
			int y = cfg_int(ae, "y", "y-coordinate of overlay", false, 0);

			filters -> push_back(new filter_overlay(s_pic, x, y));
		}
		else if (s_type == "text") {
			std::string s_text = cfg_str(ae, "text", "what text to show", false, "");
			std::string s_position = cfg_str(ae, "position", "where to put it, e.g. upper-left, lower-center, center-right, etc", false, "");

			pos_t pos = pos_to_pos(s_position);

			filters -> push_back(new filter_add_text(s_text, pos));
		}
		else if (s_type == "scaled-text") {
			std::string s_text = cfg_str(ae, "text", "what text to show", false, "");
			std::string font = cfg_str(ae, "font", "which font to use", false, "");
			int x = cfg_int(ae, "x", "x-coordinate of text", false, 0);
			int y = cfg_int(ae, "y", "y-coordinate of text", false, 0);
			int fs = cfg_int(ae, "font-size", "font size (in pixels)", false, 10);
			int r = cfg_int(ae, "r", "red component of text color", true, 0);
			int g = cfg_int(ae, "g", "green component of text color", true, 0);
			int b = cfg_int(ae, "b", "blue component of text color", true, 0);

			filters -> push_back(new filter_add_scaled_text(s_text, font, x, y, fs, r, g, b));
		}
		else if (s_type == "average") {
			int average_n = cfg_int(ae, "average-n", "how many frames to average", false, 3);

			filters -> push_back(new filter_average(average_n));
		}
		else if (s_type == "median") {
			int median_n = cfg_int(ae, "median-n", "how many frames to find the median of", false, 3);

			filters -> push_back(new filter_median(median_n));
		}
		else {
			error_exit(false, "Filter %s is not known", s_type.c_str());
		}
	}

	return filters;
}

source * load_source(configuration_t *const cfg, const Setting & o_source, resize *const r, int loglevel)
{
	source *s = nullptr;

	try
	{
		const std::string id = cfg_str(o_source, "id", "some identifier: used for selecting this module", true, "");
		log(LL_INFO, "Configuring the video-source %s...", id.c_str());

		const std::string s_type = cfg_str(o_source, "type", "source-type", false, "");
		const std::string descr = cfg_str(o_source, "descr", "description: visible in e.g. the http server", true, "");

		double max_fps = cfg_float(o_source, "max-fps", "limit the number of frames per second acquired to this value or -1.0 to disable", true, -1.0);
		if (max_fps == 0)
			error_exit(false, "Video-source: max-fps must be either > 0 or -1.0. Use -1.0 for no FPS limit.");

		int resize_w = cfg_int(o_source, "resize-width", "resize picture width to this (-1 to disable)", true, -1);
		int resize_h = cfg_int(o_source, "resize-height", "resize picture height to this (-1 to disable)", true, -1);

		double timeout = cfg_float(o_source, "timeout", "how long to wait for the camera before considering it be offline (in seconds)", true, 1.0);

		std::string exec_failure = cfg_str(o_source, "exec-failure", "script to start when the camera/video-source fails", true, "");

		std::vector<filter *> *source_filters = nullptr;
		try {
			const Setting & f = o_source.lookup("filters");
			source_filters = load_filters(cfg, f, r, nullptr);
		}
		catch(SettingNotFoundException & snfe) {
		}

		if (s_type == "v4l") {
			std::string pixel_format = cfg_str(o_source, "pixel-format", "format of the pixel-values. e.g. YUYV, JPEG, RGB3", false, "YUYV");
			bool rpi_wa = cfg_bool(o_source, "rpi-workaround", "the raspberry pi camera has a bug, this enables a workaround", true, false);
			int w = cfg_int(o_source, "width", "width of picture", false);
			int h = cfg_int(o_source, "height", "height of picture", false);
			int jpeg_quality = cfg_int(o_source, "quality", "JPEG quality, this influences the size", true, 75);
			std::string dev = cfg_str(o_source, "device", "linux v4l2 device", false, "/dev/video0");

			s = new source_v4l(id, descr, exec_failure, dev, pixel_format, rpi_wa, jpeg_quality, max_fps, w, h, r, resize_w, resize_h, loglevel, timeout, source_filters);
		}
		else if (s_type == "jpeg") {
			bool ign_cert = cfg_bool(o_source, "ignore-cert", "ignore SSL errors", true, false);
			const std::string auth = cfg_str(o_source, "http-auth", "HTTP authentication string", true, "");
			const std::string url = cfg_str(o_source, "url", "address of JPEG stream", false, "");

			s = new source_http_jpeg(id, descr, exec_failure, url, ign_cert, auth, max_fps, r, resize_w, resize_h, loglevel, timeout, source_filters);
		}
		else if (s_type == "broken-mjpeg") {
			const std::string url = cfg_str(o_source, "url", "address of MJPEG stream", false, "");
			bool ign_cert = cfg_bool(o_source, "ignore-cert", "ignore SSL errors", true, false);

			s = new source_http_mjpeg(id, descr, exec_failure, url, ign_cert, max_fps, r, resize_w, resize_h, loglevel, timeout, source_filters);
		}

		else if (s_type == "rtsp" || s_type == "mjpeg" || s_type == "stream") {
			const std::string url = cfg_str(o_source, "url", "address of video stream", false, "");
			bool tcp = cfg_bool(o_source, "tcp", "use TCP for RTSP transport (instead of default UDP)", true, false);

			s = new source_stream(id, descr, exec_failure, url, tcp, max_fps, r, resize_w, resize_h, loglevel, timeout, source_filters);
		}
		else if (s_type == "plugin") {
			std::string plugin_bin = cfg_str(o_source, "source-plugin-file", "filename of video data source plugin", true, "");
			std::string plugin_arg = cfg_str(o_source, "source-plugin-parameter", "parameter for video data source plugin", true, "");

			s = new source_plugin(id, descr, exec_failure, plugin_bin, plugin_arg, max_fps, r, resize_w, resize_h, loglevel, timeout, source_filters);
		}
		else if (s_type == "delay") {
			int n_frames = cfg_int(o_source, "n-frames", "how many frames in the past", false, 1);
			int jpeg_quality = cfg_int(o_source, "quality", "JPEG quality, this influences the size", true, 75);
			std::string delayed_source = cfg_str(o_source, "delayed-source", "source to delay", false, "");

			source *s_s = (source *)find_interface_by_id(cfg, delayed_source);
			if (!s_s)
				error_exit(false, "Source \"%s\" is not known", delayed_source.c_str());

			s = new source_delay(id, descr, exec_failure, s_s, jpeg_quality, n_frames, max_fps, r, resize_w, resize_h, loglevel, timeout, source_filters);
		}
		else if (s_type == "gstreamer") {
			std::string host = cfg_str(o_source, "host", "IPv4 address to send to", false, "");
			int port = cfg_int(o_source, "port", "port to send to", false, 9000);

			s = new source_gstreamer(id, descr, exec_failure, host, port, max_fps, r, resize_w, resize_h, loglevel, timeout, source_filters);
		}
		else if (s_type == "pixelflood") {
			int listen_port = cfg_int(o_source, "listen-port", "Port to listen on", false, 5004);
			int pixel_size = cfg_int(o_source, "pixel-size", "Pixel-size", true, 1);
			std::string pp_str = cfg_str(o_source, "pp", "Pixelflut format: tcp-txt or udp-bin", false, "tcp-txt");
			int w = cfg_int(o_source, "width", "width of picture", false);
			int h = cfg_int(o_source, "height", "height of picture", false);

			pixelflood_protocol_t pp = cfg_convert_pp(pp_str);

			s = new source_pixelflood(id, descr, exec_failure, listen_port, pixel_size, pp, max_fps, w, h, loglevel, source_filters);
		}
		else {
			error_exit(false, "Source-type \"%s\" is not known", s_type.c_str());
		}
	}
	catch(SettingException & se) {
		error_exit(false, "Error in %s", se.getPath());
	}

	return s;
}

stream_plugin_t * load_stream_plugin(const Setting & in)
{
	std::string file = cfg_str(in, "stream-writer-plugin-file", "filename of stream writer plugin", false, "");
	if (file.empty())
		return nullptr;

	stream_plugin_t *sp = new stream_plugin_t;

	sp -> par = cfg_str(in, "stream-writer-plugin-parameter", "parameter for stream writer plugin", true, "");

	void *library = dlopen(file.c_str(), RTLD_NOW);
	if (!library)
		error_exit(true, "Failed opening stream writer library %s", file.c_str());

	sp -> init_plugin = (tp_init_plugin_t)find_symbol(library, "init_plugin", "stream writer plugin", file.c_str());
	sp -> open_file = (open_file_t)find_symbol(library, "open_file", "stream writer plugin", file.c_str());
	sp -> write_frame = (write_frame_t)find_symbol(library, "write_frame", "stream writer plugin", file.c_str());
	sp -> close_file = (close_file_t)find_symbol(library, "close_file", "stream writer plugin", file.c_str());
	sp -> uninit_plugin = (tp_uninit_plugin_t)find_symbol(library, "uninit_plugin", "stream writer plugin", file.c_str());

	return sp;
}

void interval_fps_error(const char *const name, const char *what, const char *id)
{
	error_exit(false, "Interval/%s %s not set or invalid (e.g. 0) for target (%s). Make sure that you use a float-value for the fps/interval, e.g. 13.0 instead of 13", name, what, id);
}

target * load_target(const Setting & in, source *const s, resize *const r, meta *const m, configuration_t *const cfg)
{
	target *t = nullptr;

	const std::string id = cfg_str(in, "id", "some identifier: used for selecting this module", true, "");
	const std::string descr = cfg_str(in, "descr", "description: visible in e.g. the http server", true, "");

	log(LL_INFO, "Loading target %s", id.c_str());

#ifdef WITH_GWAVI
	std::string format = cfg_str(in, "format", "avi, extpipe, ffmpeg (for mp4, flv, etc), jpeg, vnc, plugin, as-a-new-source or pixelflood", false, "");
#else
	std::string format = cfg_str(in, "format", "extpipe, ffmpeg (for mp4, flv, etc), jpeg, vnc, plugin, as-a-new-source or pixelflood", false, "");
#endif

	std::string path = cfg_str(in, "path", "directory to write to", format == "as-a-new-source", "");
	std::string prefix = cfg_str(in, "prefix", "string to begin filename with", true, "");
	std::string fmt = cfg_str(in, "fmt", "specifies filename format", true, default_fmt);

	std::string exec_start = cfg_str(in, "exec-start", "script to start when recording starts", true, "");
	std::string exec_cycle = cfg_str(in, "exec-cycle", "script to start when the record file is restarted", true, "");
	std::string exec_end = cfg_str(in, "exec-end", "script to start when the recording stops", true, "");

	int restart_interval = cfg_int(in, "restart-interval", "after how many seconds should the stream-file be restarted", true, -1);

	int jpeg_quality = cfg_int(in, "quality", "JPEG quality, this influences the size", true, 75);

	bool use_previous_when_fail = cfg_bool(in, "use-previous-when-fail", "use previous frame when current is not available", true, false);

	double fps = 0, interval = 0;
	if (!find_interval_or_fps(in, &interval, "fps", &fps))
		interval_fps_error("fps", "for writing frames to disk (FPS as the video frames are retrieved)", id.c_str());

	double override_fps = cfg_float(in, "override-fps", "What FPS to use in the output file. That way you can let the resulting file be played faster (or slower) than how the stream is obtained from the camera. -1.0 to disable", true, -1.0);

	std::vector<filter *> *filters = nullptr;
	try {
		const Setting & f = in.lookup("filters");
		filters = load_filters(cfg, f, r, m);
	}
	catch(SettingNotFoundException & snfe) {
	}

	if (format == "avi")
#ifdef WITH_GWAVI
		t = new target_avi(id, descr, s, path, prefix, fmt, jpeg_quality, restart_interval, interval, filters, exec_start, exec_cycle, exec_end, override_fps, cfg, false, use_previous_when_fail);
#else
		goto ffmpeg;  // yes
#endif
	else if (format == "extpipe") {
		std::string cmd = cfg_str(in, "cmd", "Command to send the frames to", false, "");

		t = new target_extpipe(id, descr, s, path, prefix, fmt, jpeg_quality, restart_interval, interval, filters, exec_start, exec_cycle, exec_end, cmd, cfg, false, use_previous_when_fail);
	}
	else if (format == "ffmpeg") {
	ffmpeg:
		int bitrate = cfg_int(in, "bitrate", "How many bits per second to emit. For 352x288 200000 is a sane value. This value affects the quality.", true, 200000);
		std::string type = cfg_str(in, "ffmpeg-type", "E.g. flv, mp4", true, "mp4");
		std::string const parameters = cfg_str(in, "ffmpeg-parameters", "Parameters specific for ffmpeg.", true, "");

		t = new target_ffmpeg(id, descr, parameters, s, path, prefix, fmt, restart_interval, interval, type, bitrate, filters, exec_start, exec_cycle, exec_end, override_fps, cfg, false, use_previous_when_fail);
	}
	else if (format == "jpeg")
		t = new target_jpeg(id, descr, s, path, prefix, fmt, jpeg_quality, restart_interval, interval, filters, exec_start, exec_cycle, exec_end, cfg, false, use_previous_when_fail);
	else if (format == "plugin") {
		stream_plugin_t *sp = load_stream_plugin(in);

		t = new target_plugin(id, descr, s, path, prefix, fmt, jpeg_quality, restart_interval, interval, filters, exec_start, exec_cycle, exec_end, sp, override_fps, cfg, false, use_previous_when_fail);
	}
	else if (format == "vnc") {
		std::string listen_adapter = cfg_str(in, "listen-adapter", "network interface to listen on or 0.0.0.0 for all", false, "");
		int listen_port = cfg_int(in, "listen-port", "port to listen on", false, 5901);

		t = new target_vnc(id, descr, s, listen_adapter, listen_port, restart_interval, interval, filters, exec_start, exec_end, cfg, false, use_previous_when_fail);
	}
	else if (format == "pixelflood") {
		std::string host = cfg_str(in, "host", "IPv4 address to send to", false, "");
		int port = cfg_int(in, "port", "port to send to", false, 5004);
		int pfw = cfg_int(in, "pf-w", "width of pixelflood target", true, 128);
		int pfh = cfg_int(in, "pf-h", "height of pixelflood target", true, 16);
		int quality = 100;
		std::string pp_str = cfg_str(in, "pp", "Pixelflut format: tcp-txt or udp-bin", false, "tcp-txt");
		int xoff = cfg_int(in, "x-off", "x-offset in pixelflood target", true, 0);
		int yoff = cfg_int(in, "y-off", "y-offset in pixelflood target", true, 0);

		pixelflood_protocol_t pp = cfg_convert_pp(pp_str);

		t = new target_pixelflood(id, descr, s, interval, filters, -1, cfg, host, port, r, pfw, pfh, quality, pp, xoff, yoff, use_previous_when_fail);
	}
	else if (format == "as-a-new-source") {
		const std::string new_id = cfg_str(in, "new-id", "some identifier: used for selecting this module", false, "");
		const std::string new_descr = cfg_str(in, "new-descr", "description: visible in e.g. the http server", false, "");

		t = new target_new_source(id, descr, s, interval, filters, cfg, new_id, new_descr);

		instance_t *new_source_instance = new instance_t;
		new_source_instance->name = myformat("%s to %s", id.c_str(), new_id.c_str());
		new_source_instance->interfaces.push_back(((target_new_source *)t) -> get_new_source());

		cfg->instances.push_back(new_source_instance);
	}
	else {
		error_exit(false, "Format %s is unknown (stream to disk backends)", format.c_str());
	}

	return t;
}

std::vector<interface *> load_motion_triggers(configuration_t *const cfg, const Setting &o_mt, source *const s, resize *const r, instance_t *const ci)
{
	std::vector<interface *> out;

	size_t n_vlb = o_mt.getLength();

	log(LL_DEBUG, " %zu motion trigger(s)", n_vlb);

	if (n_vlb == 0)
		log(LL_INFO, " 0 triggers, is that correct?");

	for(size_t i=0; i<n_vlb; i++) {
		const Setting &trigger = o_mt[i];

		const std::string id = cfg_str(trigger, "id", "some identifier: used for selecting this module", true, "");
		const std::string descr = cfg_str(trigger, "descr", "description: visible in e.g. the http server", true, "");

		std::map<std::string, parameter *> dp; // detection parameters

		/////////////
		std::vector<target *> *motion_targets = new std::vector<target *>();

		try {
			const Setting & t = trigger.lookup("targets");

			size_t n_t = t.getLength();
			log(LL_DEBUG, " %zu motion trigger target(s)", n_t);

			if (n_t == 0)
				log(LL_WARNING, " 0 trigger targets, is that correct?");

			for(size_t i=0; i<n_t; i++) {
				const Setting & ct = t[i];

				motion_targets -> push_back(load_target(ct, s, r, s -> get_meta(), cfg));
			}
		}
		catch(SettingNotFoundException & snfe) {
		}
		//////////

		std::string motion_source = cfg_str(trigger, "motion-source", "source to monitor, leave empty for monitoring by itself (less efficient)", true, "");
		instance_t *const lr_motion_trigger = motion_source.empty() ? nullptr : (instance_t *)find_instance_by_name(cfg, motion_source);

		interface *m = nullptr;

		if (i) {
			m = new motion_trigger_other_source(id, descr, s, lr_motion_trigger, motion_targets, dp);
		}
		else {
			cfg_int(trigger, "noise-factor", "at what difference levell is the pixel considered to be changed", true, 32, &dp);
			cfg_float(trigger, "min-pixels-changed-percentage", "minimum percentage (%%) of pixels need to be changed before the motion trigger is triggered", true, 1.0, &dp);
			cfg_float(trigger, "max-pixels-changed-percentage", "maximum percentage (%%) of pixels need to be changed before the motion trigger is triggered", true, 100.0, &dp);

			cfg_int(trigger, "min-duration", "minimum number of frames to record", true, 5, &dp);
			cfg_int(trigger, "mute-duration", "how long not to record (in frames) after motion has stopped", true, 5, &dp);
			cfg_int(trigger, "min-n-frames", "how many frames should have motion before recording starts", true, 1, &dp);
			cfg_int(trigger, "pre-motion-record-duration", "how many frames to record that happened before the motion started", true, 10, &dp);
			int warmup_duration = cfg_int(trigger, "warmup-duration", "how many frames to ignore so that the camera can warm-up", true, 10);

			if (cfg_float(trigger, "max-fps", "maximum number of frames per second to analyze (or -1.0 for no limit)", true, -1.0, &dp) == 0)
				error_exit(false, "Motion triggers: max-fps must be either > 0 or -1.0. Use -1.0 for no FPS limit.");

			std::string selection_bitmap = cfg_str(trigger, "selection-bitmap", "bitmaps indicating which pixels to look at. must be same size as webcam image and must be a .pbm-file. leave empty to disable.", true, "");
			selection_mask *sm = selection_bitmap.empty() ? nullptr : new selection_mask(r, selection_bitmap);

			std::vector<filter *> *filters_detection = nullptr;
			try {
				const Setting & f = trigger.lookup("filters-detection");
				filters_detection = load_filters(cfg, f, r, s -> get_meta());
			}
			catch(SettingNotFoundException & snfe) {
			}

			std::string file = cfg_str(trigger, "trigger-plugin-file", "filename of motion detection plugin", true, "");
			ext_trigger_t *et = nullptr;
			if (!file.empty()) {
				et = new ext_trigger_t;
				et -> par = cfg_str(trigger, "trigger-plugin-parameter", "parameter for motion detection plugin", true, "");
				et -> library = dlopen(file.c_str(), RTLD_NOW);
				if (!et -> library)
					error_exit(true, "Failed opening motion detection plugin library %s", file.c_str());

				et -> init_motion_trigger = (init_motion_trigger_t)find_symbol(et -> library, "init_motion_trigger", "motion detection plugin", file.c_str());
				et -> detect_motion = (detect_motion_t)find_symbol(et -> library, "detect_motion", "motion detection plugin", file.c_str());
				et -> uninit_motion_trigger = (uninit_motion_trigger_t)find_symbol(et -> library, "uninit_motion_trigger", "motion detection plugin", file.c_str());
			}

			std::string exec_start = cfg_str(trigger, "exec-start", "script to start when recording starts", true, "");
			std::string exec_end = cfg_str(trigger, "exec-end", "script to start when the recording stops", true, "");

			m = new motion_trigger_generic(id, descr, s, warmup_duration, filters_detection, motion_targets, sm, et, exec_start, exec_end, ci, dp);
		}

		out.push_back(m);
	}

	return out;
}

view * load_view(configuration_t *const cfg, const Setting & in, resize *const r, instance_t *const inst)
{
	const std::string id = cfg_str(in, "id", "some identifier: used for selecting this module", true, "");
	const std::string descr = cfg_str(in, "descr", "description: visible in e.g. the http server", true, "");

	const std::string type = cfg_str(in, "type", "type: html-grid", false, "html-grid");

	std::vector<view_src_t> ids;
	const Setting & s_ids = in.lookup("sources");

	for(int i=0; i<s_ids.getLength(); i++) {
		const Setting & src = s_ids[i];

		std::string id = cfg_str(src, "id", "ID of source", false, "");

		std::string p = cfg_str(src, "position", "where to put the source", true, "none");
		pos_t pos = pos_to_pos(p);

		ids.push_back({ id, pos });
	}

	int width = cfg_int(in, "width", "width of the total view", true, -1);
	int height = cfg_int(in, "height", "height of the total view", true, -1);

	int gwidth = cfg_int(in, "grid-width", "number of columns", true, -1);
	int gheight = cfg_int(in, "grid-height", "number of rows", true, -1);

	double interval = cfg_float(in, "switch-interval", "when to switch source", true, 5.0);

	view *v = nullptr;

	if (type == "html-grid")
		v = new view_html_grid(cfg, id, descr, width, height, ids, gwidth, gheight, interval);
	else if (type == "source-switch" || type == "all" || type == "pip") {
		std::vector<target *> *targets = new std::vector<target *>();

		std::vector<filter *> *filters = nullptr;
		try {
			const Setting & f = in.lookup("filters");
			filters = load_filters(cfg, f, r, nullptr);
		}
		catch(SettingNotFoundException & snfe) {
		}

		if (type == "source-switch") {
			bool auto_shrink = cfg_bool(in, "auto-shrink", "automatically resize when source-size does not match selected w/h of view", true, true);

			v = new view_ss(cfg, id, descr, width, height, auto_shrink, r, ids, interval, targets, filters);
		}
		else if (type == "all") {
			if (width == -1 || height == -1)
				error_exit(false, "view of type \"all\" must have a valid width/height, not -1");

			if (gwidth == -1 || gheight == -1)
				error_exit(false, "view of type \"all\" must have a valid grid-width/grid-height, not -1");

			v = new view_all(cfg, id, descr, width, height, gwidth, gheight, r, ids, interval, targets, filters);
		}
		else if (type == "pip") {
			int perc = cfg_int(in, "percentage", "resize (shrink) percentage", false, 25);

			v = new view_pip(cfg, id, descr, width, height, r, ids, targets, filters, perc);
		}

		try {
			const Setting & t = in.lookup("targets");

			size_t n_t = t.getLength();
			log(LL_DEBUG, " %zu view target(s) for %s", n_t, id.c_str());

			for(size_t i=0; i<n_t; i++) {
				const Setting & ct = t[i];

				targets -> push_back(load_target(ct, v, r, nullptr, cfg));
			}
		}
		catch(SettingNotFoundException & snfe) {
		}
	}
	else if (type == "html-all") {
		ids.clear();

		for(instance_t * cur_inst : cfg -> instances) {
			source *s = find_source(cur_inst);
			ids.push_back({ s -> get_id(), none });
		}

		v = new view_html_all(cfg, id, descr, ids);
	}
	else
		error_exit(false, "Currently only \"html-grid\", \"source-switch\", \"view-all\" and \"html-all\" (for views, id %s) are supported", id.c_str());

	return v;
}

std::vector<interface *> load_http_servers(configuration_t *const cfg, instance_t *const ci, const Setting & hs, const bool local_instance, source *const s, resize *const r, instance_t *const views, meta *const m)
{
	size_t n_hl = hs.getLength();

	log(LL_DEBUG, " %zu http server(s)", n_hl);

	std::vector<interface *> out;

	if (n_hl == 0 && ci)
		log(LL_WARNING, " 0 servers, is that correct?");

	for(size_t i=0; i<n_hl; i++) {
		const Setting &server = hs[i];

		const std::string id = cfg_str(server, "id", "some identifier: used for selecting this module", true, "");
		const std::string descr = cfg_str(server, "descr", "description: visible in e.g. the http server", true, "");

		std::string listen_adapter = cfg_str(server, "listen-adapter", "network interface to listen on or 0.0.0.0 for all", false, "");
		int listen_port = cfg_int(server, "listen-port", "port to listen on", false, 8080);
		log(LL_INFO, " HTTP server listening on %s:%d", listen_adapter.c_str(), listen_port);

		std::string auth_method = cfg_str(server, "http-auth-method", "none, file or pam", true, "none");

		http_auth *auth = nullptr;
		if (auth_method == "file") {
			std::string auth_file = cfg_str(server, "http-auth-file", "file containing authentication data", true, "");
			auth = new http_auth(auth_file);
		}
		else if (auth_method == "pam") {
			auth = new http_auth_pam();
		}
		else if (auth_method != "none") {
			error_exit(false, "Authentication method \"%s\" not understood", auth_method.c_str());
		}

		int jpeg_quality = cfg_int(server, "quality", "JPEG quality, this influences the size", true, 75);
		int time_limit = cfg_int(server, "time-limit", "how long (in seconds) to stream before the connection is closed", true, -1);
		bool use_previous_when_fail = cfg_bool(server, "use-previous-when-fail", "use previous frame when current is not available", true, false);

		int resize_w = cfg_int(server, "resize-width", "resize picture width to this (-1 to disable)", true, -1);
		int resize_h = cfg_int(server, "resize-height", "resize picture height to this (-1 to disable)", true, -1);

		bool motion_compatible = cfg_bool(server, "motion-compatible", "only stream MJPEG and do not wait for HTTP request", true, false);
		bool allow_admin = cfg_bool(server, "allow-admin", "when enabled, you can partially configure services", true, false);
		bool is_rest = cfg_bool(server, "is-rest", "when enabled then this is a REST (only!) interface", true, false);
		bool archive_access = cfg_bool(server, "archive-access", "when enabled, you can retrieve recorded video/images", true, false);

		std::string snapshot_dir = cfg_str(server, "snapshot-dir", "where to store snapshots (triggered by HTTP server). see \"allow-admin\".", true, "");

		std::vector<filter *> *http_filters = nullptr;
		try {
			const Setting & f = server.lookup("filters");
			http_filters = load_filters(cfg, f, r, m);
		}
		catch(SettingNotFoundException & snfe) {
		}

		double fps = 0, interval = 0;
		if (!find_interval_or_fps(server, &interval, "fps", &fps) && is_rest == false)
			interval_fps_error("fps", "for showing video frames", id.c_str());

		interface *h = new http_server(cfg, auth, ci, id, descr, listen_adapter, listen_port, fps, jpeg_quality, time_limit, http_filters, r, resize_w, resize_h, motion_compatible ? s : nullptr, allow_admin, archive_access, snapshot_dir, is_rest, views, use_previous_when_fail);

		if (ci)
			ci -> interfaces.push_back(h);

		out.push_back(h);
	}

	return out;
}

interface * load_v4l2_loopback(configuration_t *const cfg, const Setting & o_vlb, source *const s, instance_t *const ci)
{
	const std::string id = cfg_str(o_vlb, "id", "some identifier: used for selecting this module", true, "");
	const std::string descr = cfg_str(o_vlb, "descr", "description: visible in e.g. the http server", true, "");
	const std::string pixel_format = cfg_str(o_vlb, "pixel-format", "format of the pixel-values. YUV420, RGB24", false, "YUV420");

	std::string dev = cfg_str(o_vlb, "device", "Linux v4l2 device to connect to", true, "");

	double fps = 0, interval = 0;
	if (!find_interval_or_fps(o_vlb, &interval, "fps", &fps))
		interval_fps_error("fps", "for sending frames to loopback", id.c_str());

	std::vector<filter *> *filters = nullptr;
	try {
		const Setting & f = o_vlb.lookup("filters");
		filters = load_filters(cfg, f, cfg->r, s -> get_meta());
	}
	catch(SettingNotFoundException & snfe) {
	}

	return new v4l2_loopback(id, descr, s, fps, dev, pixel_format, filters, ci);
}

void load_configuration(configuration_t *const cfg, const bool verbose, const int ll)
{
	cfg->lock.lock();

	log(LL_INFO, "Loading %s...", cfg->cfg_file.c_str());

	Config lc_cfg;
	try
	{
		lc_cfg.readFile(cfg->cfg_file.c_str());
	}
	catch(const FileIOException &fioex)
	{
		error_exit(true, "I/O error while reading configuration file %s", cfg->cfg_file.c_str());
	}
	catch(const ParseException &pex)
	{
		error_exit(true, "Configuration file %s parse error at line %d: %s", pex.getFile(), pex.getLine(), pex.getError());
	}

	const Setting& root = lc_cfg.getRoot();

	//***

	std::string logfile = cfg_str(lc_cfg, "logfile", "file where to store logging", false, "");

	int loglevel = -1;

	if (verbose)
		loglevel = ll;
	else {
		std::string level = cfg_str(lc_cfg, "log-level", "log level (fatal, error, warning, info, debug, debug-verbose)", false, "");
		const char *ll = level.c_str();
		if (strcasecmp(ll, "fatal") == 0)
			loglevel = LL_FATAL;
		else if (strcasecmp(ll, "error") == 0)
			loglevel = LL_ERR;
		else if (strcasecmp(ll, "warning") == 0)
			loglevel = LL_WARNING;
		else if (strcasecmp(ll, "info") == 0)
			loglevel = LL_INFO;
		else if (strcasecmp(ll, "debug") == 0)
			loglevel = LL_DEBUG;
		else if (strcasecmp(ll, "debug-verbose") == 0)
			loglevel = LL_DEBUG_VERBOSE;
		else
			error_exit(false, "Log level '%s' not recognized", ll);

	}

	setlogfile(logfile.empty() ? nullptr : logfile.c_str(), loglevel);

	log(LL_INFO, " *** " NAME " v" VERSION " starting ***");

	std::string resize_type = cfg_str(lc_cfg, "resize-type", "can be regular, crop or cairo. This selects what method will be used for resizing the video stream (if requested).", true, "");
	bool resize_crop_center = cfg_bool(lc_cfg, "resize-crop-center", "center after crop", true, false);

	cfg->r = nullptr;
	if (resize_type == "cairo")
		cfg->r = new resize_cairo();
	else if (resize_type == "regular" || resize_type == "") {
		cfg->r = new resize();

		if (resize_type == "")
			log(LL_INFO, "No resizer/scaler selected (\"resize-type\"); assuming default");
	}
	else if (resize_type == "fine")
		cfg->r = new resize_fine();
	else if (resize_type == "crop")
		cfg->r = new resize_crop(resize_crop_center);
	else {
		error_exit(false, "Scaler/resizer of type \"%s\" is not known", resize_type.c_str());
	}

	log(LL_INFO, "Configuring maintenance settings...");
	cfg->clnr = nullptr;
	try {
		const Setting & m = root.lookup("maintenance");

		cfg->db_file = cfg_str(m, "db-file", "database file", true, "");
		db *dbi = register_database(cfg->db_file);

		int clean_interval = cfg_int(m, "clean-interval", "how often (in seconds) to check for old files (-1 to disable)", true, -1);
		int keep_max = cfg_int(m, "keep-max", "after how many days should things be deleted (-1 to disable)", true, -1);
		cfg->clnr = new cleaner(dbi, clean_interval, keep_max * 86400);
	}
	catch(SettingNotFoundException & snfe) {
		register_database("");
		log(LL_INFO, " no maintenance-items set");
	}

	const Setting &o_instances = root["instances"];
	size_t n_instances = o_instances.getLength();

	if (n_instances == 0)
		log(LL_WARNING, " 0 instances, is that correct? Note that since version 2.0 the configuration file layout changed a tiny bit to accomodate multiple instances!");

	instance_t *main_instance = nullptr;

	for(size_t i_loop=0; i_loop<n_instances; i_loop++) {
		const Setting &instance_root = o_instances[i_loop];
		instance_t *ci = new instance_t; // ci = current instance

		if (main_instance == nullptr)
			main_instance = ci;

		ci -> name = cfg_str(instance_root, "instance-name", "instance-name", true, "default instance name");

		//***
		source *s = load_source(cfg, instance_root["source"], cfg->r, loglevel);
		ci -> interfaces.push_back(s);

		//***
		log(LL_INFO, "Configuring the meta sources...");

		try {
			const Setting &ms = instance_root["meta-plugin"];
			size_t n_ms = ms.getLength();

			log(LL_DEBUG, " %zu meta plugin(s)", n_ms);

			if (n_ms == 0)
				log(LL_WARNING, " 0 plugins, is that correct?");

			for(size_t i=0; i<n_ms; i++) {
				const Setting &server = ms[i];

				std::string file = cfg_str(server, "plugin-file", "filename of the meta plugin", false, "");
				std::string par = cfg_str(server, "plugin-parameter", "parameter of the meta plugin", true, "");

				s -> get_meta() -> add_plugin(file, par.c_str());
			}
		}
		catch(const SettingNotFoundException &nfex) {
			log(LL_INFO, " no meta source");
		}

		// listen adapter, listen port, source, fps, jpeg quality, time limit (in seconds)
		log(LL_INFO, "Configuring the HTTP server(s)...");

		try {
			load_http_servers(cfg, ci, instance_root["http-server"], true, s, cfg->r, nullptr, s -> get_meta());
		}
		catch(const SettingNotFoundException &nfex) {
			log(LL_INFO, " no HTTP server");
		}
		//	catch(SettingException & se) {
		//		error_exit(false, "Error in %s", se.getPath());
		//	}

		//***

		log(LL_INFO, "Configuring a video-loopback...");
		try
		{
			const Setting &o_vlb = instance_root["video-loopback"];

			interface *f = load_v4l2_loopback(cfg, o_vlb, s, ci);
			ci -> interfaces.push_back(f);
		}
		catch(const SettingNotFoundException &nfex) {
			log(LL_INFO, " no video loopback");
		}

		//***

		log(LL_INFO, "Configuring the motion trigger(s)...");
		try
		{
			std::vector<interface *> triggers = load_motion_triggers(cfg, instance_root["motion-trigger"], s, cfg->r, ci);

			ci->interfaces.insert(ci->interfaces.end(), triggers.begin(), triggers.end());
		}
		catch(const SettingNotFoundException &nfex) {
			log(LL_INFO, " no motion trigger defined");
		}

		//***
		log(LL_INFO, "Configuring the stream-to-disk backend(s)...");
		try {
			const Setting & t = instance_root.lookup("stream-to-disk");

			size_t n_t = t.getLength();
			log(LL_DEBUG, " %zu disk streams", n_t);

			if (n_t == 0)
				log(LL_WARNING, " 0 streams, is that correct?");

			for(size_t i=0; i<n_t; i++) {
				const Setting & ct = t[i];

				interface *tgt = load_target(ct, s, cfg->r, s -> get_meta(), cfg);
				ci -> interfaces.push_back(tgt);
			}
		}
		catch(SettingNotFoundException & snfe) {
			log(LL_INFO, " no stream-to-disk backends defined");
		}

		if (!s)
			error_exit(false, "No video-source selected");

		cfg->instances.push_back(ci);
	}

	//***
	log(LL_INFO, "Configuring views...");
	instance_t *views = nullptr;
	try {
		const Setting & v = root.lookup("views");

		views = new instance_t; // ci = current instance
		views -> name = "views";

		size_t n_v = v.getLength();
		log(LL_DEBUG, " %zu views", n_v);

		for(size_t i=0; i<n_v; i++) {
			const Setting & cv = v[i];

			view *v = load_view(cfg, cv, cfg->r, views);
			views -> interfaces.push_back(v);

			try {
				const Setting &o_vlb = cv["video-loopback"];
				log(LL_INFO, "Configuring a video-loopback...");

				interface *f = load_v4l2_loopback(cfg, o_vlb, v, views);
				views -> interfaces.push_back(f);
			}
			catch(const SettingNotFoundException &nfex) {
			}
		}

		cfg->instances.push_back(views);
	}
	catch(SettingNotFoundException & snfe) {
		log(LL_INFO, " no stream-to-disk backends defined");
	}

	try {
		log(LL_INFO, "Loading global HTTP(/REST) server(s)");
		cfg->global_http_servers = load_http_servers(cfg, nullptr, root["global-http-server"], false, nullptr, nullptr, views, nullptr);
	}
	catch(const SettingNotFoundException &nfex) {
		log(LL_INFO, " no global HTTP(/REST) server(s)");
	}

	std::set<std::string> check_id;
	for(instance_t * inst : cfg->instances) {
		for(const interface *const i: inst -> interfaces) {
			std::string cur = i -> get_id();
			//printf("%s %s %s\n", inst -> name.c_str(), cur.c_str(), i -> get_descr().c_str());

			if (check_id.find(cur) != check_id.end())
				log(LL_WARNING, "There are multiple modules with the same ID (%s): this will give problems when trying to use the REST interface!", cur.c_str());
			else
				check_id.insert(cur);
		}
	}

	check_id.clear();
	for(instance_t * inst : cfg->instances) {
		if (check_id.find(inst -> name) != check_id.end())
			log(LL_WARNING, "There are multiple instances with the same ID (%s): this will give problems when trying to use the REST and WEB-interfaces!", inst -> name.c_str());
		else
			check_id.insert(inst -> name);
	}

	cfg->lock.unlock();
}
